import ast
import functools
import inspect
import sys
import threading
import types
from datetime import datetime
from typing import Any, Callable, List, Mapping, MutableMapping, Optional, Type, TypeVar, cast

from chalk.features.feature import Feature, Features, FeatureSetBase, FeatureTypeAnnotation, FeatureWrapper
from chalk.features.tag import Tags
from chalk.utils.duration import Duration
from chalk.utils.metaprogramming import MISSING, create_fn, field_assign, set_new_attribute
from chalk.utils.string import to_snake_case

T = TypeVar("T")
U = TypeVar("U")
JsonValue = TypeVar("JsonValue")


def feature(
    description: Optional[str] = None,
    owner: Optional[str] = None,
    tags: Optional[Tags] = None,
    name: Optional[str] = None,
    version: Optional[int] = None,
    primary: bool = False,
    max_staleness: Optional[Duration] = None,
    etl_offline_to_online: bool = False,
    encoder: Optional[Callable[[T], JsonValue]] = None,
    decoder: Optional[Callable[[JsonValue], T]] = None,
) -> T:
    return cast(
        T,
        Feature(
            name=name,
            version=version,
            owner=owner,
            tags=tags,
            description=description,
            primary=primary,
            max_staleness=max_staleness,
            etl_offline_to_online=etl_offline_to_online,
            encoder=encoder,
            decoder=decoder,
        ),
    )


def feature_time() -> Any:
    return Feature(typ=datetime, is_feature_time=True)


def has_one(f: Callable[[], Any]) -> Any:
    return Feature(join=f)


def has_many(f: Callable[[], Any]) -> Any:
    return Feature(join=f)


def _discover_feature(cls: Type[Features], name: str, *conditions: Callable[[Feature], bool]):
    for cond in conditions:
        features = [c for c in cls.features if isinstance(c, Feature) and cond(c)]
        if len(features) == 1:
            return features[0]
        if len(features) > 1:
            raise ValueError(
                f"Multiple {name} features are not supported in {cls.__name__}: "
                + ", ".join(f"{cls.__name__}.{x.name}" for x in features)
            )
    return None


# @dataclass_transform(field_specifiers=(feature, feature_time, has_many, has_one))
def features(
    cls: Optional[Type[T]] = None,
    /,
    owner: Optional[str] = None,
    tags: Optional[Tags] = None,
) -> Type[T]:
    """Returns the same class as was passed in, with dunder methods
    added based on the fields defined in the class.

    Examines PEP 526 __annotations__ to determine fields.
    """

    def wrap(c: Type[T]) -> Type[T]:
        updated_class = _process_class(c, owner, tags)
        set_new_attribute(
            cls=updated_class,
            name="namespace",
            value=to_snake_case(updated_class.__name__),
        )
        assert issubclass(updated_class, Features)
        FeatureSetBase.registry[updated_class.namespace] = updated_class

        primary_feature = _discover_feature(
            updated_class,
            "primary",
            lambda f: f.primary,
            lambda f: f.name == "id" and not f.has_resolved_join and not f.is_feature_time,
        )
        if primary_feature is not None:
            primary_feature.primary = True

        timestamp_feature = _discover_feature(
            updated_class,
            "feature time",
            lambda f: f.is_feature_time,
            lambda f: f.name == "ts" and not f.has_resolved_join,
        )
        set_new_attribute(cls=updated_class, name="__chalk_primary__", value=primary_feature)
        set_new_attribute(cls=updated_class, name="__chalk_ts__", value=timestamp_feature)

        return cast(Type[T], updated_class)

    # See if we're being called as @features or @features().
    if cls is None:
        # We're called with parens.
        return wrap

    # We're called as @features without parens.
    return wrap(cls)


def _init_param(f: Feature):
    # Return the __init__ parameter string for this field.  For
    # example, the equivalent of 'x:int=3' (except instead of 'int',
    # reference a variable set to int, and instead of '3', reference a
    # variable set to 3).
    return f"{f.name}:_type_{f.name}=MISSING"


def _init_fn(fields: List[Feature], self_name: str, _globals: Mapping[str, Any]):
    # fields contains both real fields and InitVar pseudo-fields.

    # Make sure we don't have fields without defaults following fields
    # with defaults.  This actually would be caught when exec-ing the
    # function source code, but catching it here gives a better error
    # message, and future-proofs us in case we build up the function
    # using ast.

    _locals: MutableMapping[str, Any] = {f"_type_{f.name}": f.typ and f.typ.annotation for f in fields}
    _locals.update(
        {
            "MISSING": MISSING,
            # '_HAS_DEFAULT_FACTORY': _HAS_DEFAULT_FACTORY,
        }
    )

    body_lines = []
    for f in fields:
        assert f.name is not None
        line = field_assign(f.name, f.name, self_name)
        # line is None means that this field doesn't require
        # initialization (it's a pseudo-field).  Just skip it.
        if line:
            body_lines.append(line)

    # If no body lines, use 'pass'.
    if not body_lines:
        body_lines = ["pass"]

    # Add the keyword-only args.  Because the * can only be added if
    # there's at least one keyword-only arg, there needs to be a test here
    # (instead of just concatenting the lists together).
    _init_params = [_init_param(f) for f in fields]
    return create_fn(
        "__init__",
        [self_name] + _init_params,
        body_lines,
        _locals=_locals,
        _globals=_globals,
        return_type=None,
    )


def _get_field(
    cls: Type,
    annotation_name: str,
    comments: Mapping[str, str],
    class_owner: Optional[str],
    class_tags: Optional[Tags],
):
    # Return a Field object for this field name and type.  ClassVars and
    # InitVars are also returned, but marked as such (see f._field_type).
    # default_kw_only is the value of kw_only to use if there isn't a field()
    # that defines it.

    # If the default value isn't derived from Field, then it's only a
    # normal default value.  Convert it to a Field().
    default = getattr(cls, annotation_name, MISSING)

    if isinstance(default, Feature):
        f = default
    else:
        if isinstance(default, types.MemberDescriptorType):
            # This is a field in __slots__, so it has no default value.
            default = MISSING
        f = Feature(name=annotation_name, namespace=to_snake_case(cls.__name__))

    if isinstance(f, Feature) and f.description is None and annotation_name in comments:
        f.description = comments[annotation_name]

    if isinstance(f, Feature) and f.owner is None:
        f.owner = class_owner

    if isinstance(f, Feature) and class_tags is not None:
        class_tags = [class_tags] if isinstance(class_tags, str) else class_tags
        if f.tags is None:
            f.tags = []
        elif isinstance(f.tags, str):
            f.tags = [f.tags]

        f.tags.extend(class_tags)

    # Only at this point do we know the name and the type.  Set them.
    f.namespace = to_snake_case(cls.__name__)
    if f.name is None:
        # if there is a name set by the `feature()` method, use that
        f.name = annotation_name

    f.typ = FeatureTypeAnnotation(cls, annotation_name)
    f.features_cls = cls
    f.attribute_name = annotation_name
    return f


def _recursive_repr(user_function):
    # Decorator to make a repr function return "..." for a recursive
    # call.
    repr_running = set()

    @functools.wraps(user_function)
    def wrapper(self):
        key = id(self), threading.get_ident()
        if key in repr_running:
            return "..."
        repr_running.add(key)
        try:
            result = user_function(self)
        finally:
            repr_running.discard(key)
        return result

    return wrapper


def _repr_fn(fields: List[Feature], _globals: Mapping[str, Any]):
    fn = create_fn(
        name="__repr__",
        args=["self"],
        body=[
            'return self.__class__.__qualname__ + f"('
            + ", ".join([f"{f.name}={{self.{f.name}!r}}" for f in fields])
            + ')"'
        ],
        _globals=_globals,
    )
    return _recursive_repr(fn)


def _tuple_str(obj_name, fields):
    # Return a string representing each field of obj_name as a tuple
    # member.  So, if fields is ['x', 'y'] and obj_name is "self",
    # return "(self.x,self.y)".

    # Special case for the 0-tuple.
    if not fields:
        return "()"
    # Note the trailing comma, needed if this turns out to be a 1-tuple.
    return f'({",".join([f"{obj_name}.{f.name}" for f in fields])},)'


def _cmp_fn(name, op, self_tuple, other_tuple, _globals: Mapping[str, Any]):
    # Create a comparison function.  If the fields in the object are
    # named 'x' and 'y', then self_tuple is the string
    # '(self.x,self.y)' and other_tuple is the string
    # '(other.x,other.y)'.

    return create_fn(
        name=name,
        args=["self", "other"],
        body=[
            "if other.__class__ is self.__class__:",
            f" return {self_tuple}{op}{other_tuple}",
            "return NotImplemented",
        ],
        _globals=_globals,
    )


def _eq_fn(fields: List[Feature], _globals: Mapping[str, Any]):
    self_tuple = _tuple_str("self", fields)
    other_tuple = _tuple_str("other", fields)
    return _cmp_fn("__eq__", "==", self_tuple, other_tuple, _globals=_globals)


def _iter_fn(fields: List[Feature], _globals: Mapping[str, Any]):
    return create_fn(
        "__iter__",
        args=["self"],
        body=[
            f"for __chalk_f__ in self.features:",
            f"    if getattr(self, __chalk_f__.attribute_name) != MISSING and type(__chalk_f__).__name__ == 'Feature' and not __chalk_f__.is_has_one and not __chalk_f__.is_has_many:",
            f"        yield __chalk_f__.fqn, getattr(self, __chalk_f__.attribute_name)",
        ],
    )


def _parse_annotation_comments(cls: Type[T]) -> Mapping[str, str]:
    source = inspect.getsource(cls)
    source_lines = source.splitlines()
    tree = ast.parse(source)
    comments_for_annotations = {}
    if len(tree.body) != 1:
        return comments_for_annotations

    class_def = tree.body[0]
    if isinstance(class_def, ast.ClassDef):
        for stmt in class_def.body:
            if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.target, ast.Name):
                line = stmt.lineno - 2
                comments: List[str] = []
                while line >= 0 and source_lines[line].strip().startswith("#"):
                    comment = source_lines[line].strip().strip("#").strip()
                    comments.insert(0, comment)
                    line -= 1

                if len(comments) > 0:
                    comments_for_annotations[stmt.target.id] = "\n".join(comments)

    return comments_for_annotations


def _process_class(cls: Type[T], owner: Optional[str], tags: Optional[Tags]) -> Type[T]:
    cls_annotations = cls.__dict__.get("__annotations__", {})

    if cls.__module__ in sys.modules:
        _globals = sys.modules[cls.__module__].__dict__
    else:
        _globals = {}

    # Find feature times that weren't annotated.
    for (name, member) in inspect.getmembers(cls):
        if name not in cls_annotations and isinstance(member, Feature):
            # All feature types need annotations, except for datetimes, which we can automatically infer
            if member.typ is not None and member.typ.is_parsed and issubclass(member.typ.underlying, datetime):
                cls_annotations[name] = datetime
            else:
                raise TypeError(f"Member {name} is missing an annotation")

    # If we pass this function something that isn't a class, it could raise
    try:
        comments = _parse_annotation_comments(cls)
    except:
        comments = {}

    cls_fields = [
        _get_field(
            cls=cls,
            annotation_name=name,
            comments=comments,
            class_owner=owner,
            class_tags=tags,
        )
        for name in cls_annotations
    ]

    set_new_attribute(
        cls=cls,
        name="__init__",
        value=_init_fn(
            fields=cls_fields,
            # The name to use for the "self"
            # param in __init__.  Use "self"
            # if possible.
            self_name="__dataclass_self__" if "self" in cls_fields else "self",
            _globals=_globals,
        ),
    )

    set_new_attribute(cls=cls, name="__repr__", value=_repr_fn(fields=cls_fields, _globals=_globals))

    set_new_attribute(cls=cls, name="__eq__", value=_eq_fn(fields=cls_fields, _globals=_globals))
    set_new_attribute(cls=cls, name="__hash__", value=None)
    set_new_attribute(cls=cls, name="__iter__", value=_iter_fn(fields=cls_fields, _globals=_globals))

    namespace = to_snake_case(cls.__name__)
    set_new_attribute(cls=cls, name="__chalk_namespace__", value=namespace)
    set_new_attribute(cls=cls, name="__chalk_owner__", value=owner)
    set_new_attribute(cls=cls, name="__chalk_tags__", value=tags)
    set_new_attribute(cls=cls, name="features", value=cls_fields)
    set_new_attribute(cls=cls, name="__is_features__", value=True)

    for f in cls_fields:
        assert f.attribute_name is not None
        # Wrap all class features with FeatureWrapper
        setattr(cls, f.attribute_name, FeatureWrapper(f))

    return cls



import enum
from typing import Optional, Type


def get_enum_value_type(enum_cls: Type[enum.Enum]) -> Optional[Type]:
    """Get the type of the enum value. Requires that all enum member values have the same
    type; otherwise raises a TypeError.
    Returns ``None`` if the enum has no members.

    Note: if an enum contains both float and int members, then this function will return `float`.

    Args:
        enum_cls (Type[Enum]): The enum class

    Returns:
        Optional[Type]: The type of the enum member values, or None if the enum has no members.

    Raises:
        TypeError: If ``enum_cls`` has members of heterogenous types, then a TypeError is raised.
    """
    typ = None
    for x in enum_cls:
        if typ is None:
            typ = type(x.value)
            continue
        if issubclass(type(x.value), typ):
            continue
        if issubclass(type(x.value), int) and issubclass(typ, float):
            # Effectively treat ints as a subclass of floats
            continue
        if issubclass(typ, int) and issubclass(type(x.value), float):
            # Same as above, but we encountered the int first
            typ = type(x.value)
            continue
        if issubclass(typ, type(x.value)):
            # One is a subclass of the other. In this case, swap it
            typ = type(x.value)
            continue
        raise TypeError(
            f"Enum {enum_cls.__name__} has members of heterogenous types ({typ.__name__} and {type(x.value).__name__}); all members must have the same type."
        )
    return typ



import unittest

from chalk.features import Features, features, online


@features
class HomeFeatures:
    home_id: str
    address: str
    price: int
    sq_ft: int


@online
def get_address(
    hid: HomeFeatures.home_id,
) -> HomeFeatures.address:
    return "Bridge Street" if hid == 1 else "Filbert Street"


@online
def get_home_data(
    hid: HomeFeatures.home_id,
) -> Features[HomeFeatures.price, HomeFeatures.sq_ft]:
    return HomeFeatures(
        price=200_000,
        sq_ft=2_000,
    )


class FeatureResolverCallableTestCase(unittest.TestCase):
    def test_single_output(self):
        self.assertEqual(
            get_address(2),
            "Filbert Street",
        )

    def test_multiple_output(self):
        result = get_home_data(2)
        self.assertEqual(result.price, 200_000)
        self.assertEqual(result.sq_ft, 2_000)

        self.assertNotEqual(
            result,
            HomeFeatures(
                address="hello",
                price=200_000,
                sq_ft=2_000,
            ),
        )

        self.assertEqual(
            result,
            HomeFeatures(
                price=200_000,
                sq_ft=2_000,
            ),
        )


if __name__ == "__main__":
    unittest.main()



from typing import List, Optional

from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.orm import Session as SqlAlchemySession

from chalk.features import Feature, Features
from chalk.sql.query import ChalkSqlAlchemyQuery


class ChalkSqlAlchemySession:
    _session: Optional[SqlAlchemySession]

    def __init__(self, session: Optional[SqlAlchemySession]):
        self._session = session

    def set_session(self, s: SqlAlchemySession):
        self._session = s

    def query(self, *entities, **kwargs):
        if self._session is None:
            raise ValueError("Session needs to be set via chalk_session.set_session(...)")
        transformed_entities = []
        features: List[Feature] = []
        for e in entities:
            if isinstance(e, Features):
                for f in e.features:
                    features.append(f)
                    assert f.name is not None, "The name should be defined"
                    instrumented_attribute = getattr(e, f.name)
                    if isinstance(instrumented_attribute, InstrumentedAttribute):
                        transformed_entities.append(instrumented_attribute.label(f.fqn))
            else:
                raise ValueError("Values need to be wrapped")

        return ChalkSqlAlchemyQuery(features, self._session.query(*transformed_entities, **kwargs))



from chalk.sql.integrations.bigquery import BigQuerySource
from chalk.sql.integrations.mysql import MySQLSource
from chalk.sql.integrations.redshift import RedshiftSource
from chalk.sql.integrations.snowflake import SnowflakeSource

__all__ = [
    "BigQuerySource",
    "MySQLSource",
    "RedshiftSource",
    "SnowflakeSource",
]



from sqlalchemy.engine.url import URL

from chalk.sql.base.sql_source import BaseSQLSource


class SnowflakeSource(BaseSQLSource):
    def local_engine_url(self) -> URL:
        raise ValueError("Local Snowflake client not implemented")



import tempfile
import unittest

import pandas as pd
from sqlalchemy import Column, String, create_engine
from sqlalchemy.orm import declarative_base, sessionmaker

from chalk.df.ChalkDataFrameImpl import ChalkDataFrameImpl
from chalk.features import features
from chalk.features.feature import Features
from chalk.parsed.feature import FeatureType
from chalk.sql.base.protocols import IncrementalSettings
from chalk.sql.integrations.sqlite import FileSQLite

Base = declarative_base()


class UserTable(Base):
    __tablename__ = "users_testing"
    id = Column(String, primary_key=True)
    name = Column(String)
    email = Column(String)


class AccountsTable(Base):
    __tablename__ = "accounts_testing"
    id = Column(String, primary_key=True)
    user_id = Column(String)
    title = Column(String)


@features
class UserFeatures:
    id: str
    name: str
    email: str


@features
class AccountFeatures:
    id: str
    user_id: str
    title: str


tmp = tempfile.NamedTemporaryFile(delete=False)
db = FileSQLite(tmp.name)
# db = PostgreSQLSource()


class SQLBaseTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        engine = create_engine(db.local_engine_url())
        s = sessionmaker(bind=engine)()
        UserTable.__table__.create(engine)
        AccountsTable.__table__.create(engine)

        s.add_all(
            [
                UserTable(id="u_2", name="elliot", email="elliot@chalk.ai"),
                UserTable(id="u_3", name="andy", email="andy@chalk.ai"),
                UserTable(id="u_4", name="marc", email="marc@chalk.ai"),
                AccountsTable(id="a_1", user_id="u_2", title="elliot marx"),
                AccountsTable(id="a_2", user_id="u_2", title="elliot r marx"),
            ]
        )
        s.commit()

    def test_query_all(self):
        query = (
            db.query(UserFeatures(id=UserTable.id, name=UserTable.name, email=UserTable.email))
            .filter_by(name="elliot")
            .all()
        )
        self.assertIsNone(query._incremental_settings)
        result = query.execute()
        self.assertIsInstance(result, ChalkDataFrameImpl)

    def test_query_all_incremental(self):
        query = (
            db.query(UserFeatures(id=UserTable.id, name=UserTable.name, email=UserTable.email))
            .filter_by(name="elliot")
            .all(incremental=True)
        )
        self.assertEqual(
            query._incremental_settings,
            IncrementalSettings(lookback_period=None),
        )
        result = query.execute()
        self.assertIsInstance(result, ChalkDataFrameImpl)

    def test_query_all_incremental_lookback(self):
        query = (
            db.query(UserFeatures(id=UserTable.id, name=UserTable.name, email=UserTable.email))
            .filter_by(name="elliot")
            .all(incremental=IncrementalSettings(lookback_period="1d"))
        )
        self.assertEqual(
            query._incremental_settings,
            IncrementalSettings(lookback_period="1d"),
        )
        result = query.execute()
        self.assertIsInstance(result, ChalkDataFrameImpl)

    def test_query_one(self):
        query = (
            db.query(UserFeatures(id=UserTable.id, name=UserTable.name, email=UserTable.email))
            .filter_by(name="elliot")
            .one()
        )
        result = query.execute()
        assert isinstance(result, Features)
        self.assertIsNotNone(result)
        self.assertEqual(result.email, "elliot@chalk.ai")
        self.assertEqual(result.name, "elliot")
        self.assertEqual(result.id, "u_2")

    def test_query_string_one(self):
        base_query = db.query_string(
            query="select id, email from users_testing where name=:n",
            fields=dict(email=UserFeatures.email, id=UserFeatures.id),
            args=dict(n="elliot"),
        )
        result = base_query.one().execute()
        assert isinstance(result, Features)
        self.assertEqual(result.email, "elliot@chalk.ai")
        self.assertEqual(result.id, "u_2")

    def test_query_string_all(self):
        query = db.query_string(
            query="select id, email from users_testing where name=:n",
            fields=dict(email=UserFeatures.email, id=UserFeatures.id),
            args=dict(n="elliot"),
        ).all()
        self.assertIsNone(query._incremental_settings)
        result = query.execute()
        self.assertIsInstance(result, ChalkDataFrameImpl)

    def test_query_string_all_incremental_lookback(self):
        query = db.query_string(
            query="select id, email from users_testing where name=:n",
            fields=dict(email=UserFeatures.email, id=UserFeatures.id),
            args=dict(n="elliot"),
        ).all(incremental=IncrementalSettings(lookback_period="1d"))
        self.assertEqual(
            query._incremental_settings,
            IncrementalSettings(lookback_period="1d"),
        )
        result = query.execute()
        self.assertIsInstance(result, ChalkDataFrameImpl)

    def test_query_string_all_incremental(self):
        base_query = db.query_string(
            query="select id, email from users_testing where name=:n",
            fields=dict(email=UserFeatures.email, id=UserFeatures.id),
            args=dict(n="elliot"),
        ).all(incremental=True)
        self.assertEqual(
            base_query._incremental_settings,
            IncrementalSettings(lookback_period=None),
        )
        result = base_query.execute()
        self.assertIsInstance(result, ChalkDataFrameImpl)

    def test_join(self):
        query = (
            db.query(
                AccountFeatures(
                    title=AccountsTable.title,
                    user_id=AccountsTable.user_id,
                    id=AccountsTable.id,
                )
            )
            .join(UserTable, UserTable.id == AccountsTable.user_id)
            .filter(UserTable.id == "u_2")
        )
        result = query.execute()
        pd.testing.assert_frame_equal(
            result.underlying,
            pd.DataFrame(
                data=[
                    ["a_1", "u_2", "elliot marx"],
                    ["a_2", "u_2", "elliot r marx"],
                ],
                columns=[
                    FeatureType.of(AccountFeatures.id),
                    FeatureType.of(AccountFeatures.user_id),
                    FeatureType.of(AccountFeatures.title),
                ],
            ),
        )



import unittest
from datetime import datetime

from chalk.features import Features, feature_time, features, has_one, online


@features
class HomeFeatures:
    home_id: str
    address: str
    price: int
    sq_ft: int
    homeowner: "Homeowner" = has_one(lambda: Homeowner.home_id == HomeFeatures.home_id)


@features
class Homeowner:
    fullname: str
    home_id: str
    ts: datetime = feature_time()


@online
def get_home_data(
    hid: HomeFeatures.home_id, dd: HomeFeatures.homeowner.ts
) -> Features[HomeFeatures.price, HomeFeatures.sq_ft]:
    return HomeFeatures(
        price=200_000,
        sq_ft=2_000,
    )


class FeatureResolverCallableTestCase(unittest.TestCase):
    def test_multiple_output(self):
        pass


if __name__ == "__main__":
    unittest.main()



import unittest

from chalk.features import after_all, before_all
from chalk.features.hooks import Hook


@before_all
def simple_hook_init():
    return 4


@after_all
def simple_hook_teardown():
    return 5


@after_all(environment="hello")
def env_after_all():
    return 6


@before_all(environment=["hello", "itsme"])
def env_before_all():
    return 5


class Client:
    init = False


client = Client()
client2 = Client()


@before_all(environment="my-special-env")
def setup_env():
    client.init = True


@before_all(environment="my-special-env-2")
def setup_env():
    client2.init = True


class HookTestCase(unittest.TestCase):
    def test_hooks_callable(self):
        self.assertEqual(simple_hook_init(), 4)
        self.assertEqual(simple_hook_init.environment, None)

        self.assertEqual(simple_hook_teardown(), 5)
        self.assertEqual(simple_hook_teardown.environment, None)

    def test_hook_environments(self):
        self.assertEqual(env_before_all(), 5)
        self.assertEqual(env_before_all.environment, ["hello", "itsme"])

        self.assertEqual(env_after_all(), 6)
        self.assertEqual(env_after_all.environment, ["hello"])

    def run_all(self):
        self.assertEqual(client.init, False)
        self.assertEqual(client2.init, False)
        Hook.run_all_before_all("my-special-env")
        self.assertEqual(client.init, True)
        self.assertEqual(client2.init, True)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import collections.abc
import copy
import dataclasses
import types
from datetime import datetime, timedelta
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
    get_type_hints,
)

from chalk.features.tag import Tags
from chalk.utils.collections import ensure_tuple, get_unique_item
from chalk.utils.duration import Duration

if TYPE_CHECKING:
    from chalk.features.dataframe import DataFrame
    from chalk.parsed.feature import FeatureType


U = TypeVar("U")
JsonValue = TypeVar("JsonValue")


class FeatureTypeAnnotation:
    def __init__(
        self,
        features_cls: Optional[Type[Features]] = None,
        attribute_name: Optional[str] = None,
        *,
        underlying: Optional[type] = None,
    ) -> None:
        # Either pass in the underlying -- if it is already parsed -- or pass in the feature cls and attribute name

        self._features_cls = features_cls
        self._attribute_name = attribute_name
        self._is_nullable = None
        self._is_dataframe = None
        self._is_list = None
        self._is_scalar = None
        self._underlying = None
        self._parsed_annotation = None
        if underlying is not None:
            if features_cls is not None and attribute_name is not None:
                raise ValueError("If specifying the underlying, do not specify (features_cls, attribute_name)")
            self._parse_type(underlying)
        else:
            if features_cls is None or attribute_name is None:
                raise ValueError(
                    "If not specifying the underlying, then both the (features_cls, attribute_name) must be provided"
                )
        # Store the class and attribute name so we can later use typing.get_type_hints to
        # resolve any forward references in the type annotations
        # Resolution happens lazily -- after everything is imported -- to avoid circular imports

    @property
    def is_parsed(self) -> bool:
        return self._underlying is not None

    @property
    def annotation(self) -> Union[str, type]:
        """Return the type annotation, without parsing the underlying type if it is not yet already parsed."""
        if self._parsed_annotation is not None:
            # It is already parsed. Return it.
            return self._parsed_annotation
        assert self._features_cls is not None
        assert self._attribute_name is not None
        return self._features_cls.__annotations__[self._attribute_name]

    @property
    def parsed_annotation(self) -> type:
        """The parsed type annotation. It will be parsed if needed.

        Unlike :attr:`underlying`, parsed annotation contains any container or optional types, such as
        list, dataframe, or Optional.
        """
        if self._parsed_annotation is None:
            self._parse_annotation()
        assert self._parsed_annotation is not None
        return self._parsed_annotation

    def __str__(self):
        if isinstance(self.annotation, type):
            return self.annotation.__name__
        else:
            return str(self.annotation)

    def _parse_annotation(self):
        assert self._features_cls is not None
        assert self._attribute_name is not None
        hints = get_type_hints(self._features_cls)
        parsed_annotation = hints[self._attribute_name]
        self._parse_type(parsed_annotation)

    def _parse_type(self, annotation: Optional[type]):
        from chalk.features.dataframe import DataFrame

        assert self._parsed_annotation is None, "The annotation was already parsed"
        self._parsed_annotation = annotation
        origin = get_origin(annotation)
        self._is_nullable = False
        if self._features_cls is not None and self._attribute_name is not None:
            # Return a more helpful error message, since we have context
            error_ctx = f" {self._features_cls.__name__}.{self._attribute_name}"
        else:
            error_ctx = ""
        if origin in (
            Union,
            getattr(types, "UnionType", Union),
        ):  # using getattr as UnionType was introduced in python 3.10
            args = get_args(annotation)
            # If it's a union, then the only supported union is for nullable features. Validate this
            if len(args) != 2 or (None not in args and type(None) not in args):
                raise TypeError(
                    f"Invalid annotation for feature{error_ctx}: Unions with non-None types are not allowed"
                )
            annotation = args[0] if args[1] in (None, type(None)) else args[1]
            self._is_nullable = True

        # The only allowed collections here are List or DataFrame
        self._is_list = False
        if origin in (list, List):
            args = get_args(annotation)
            assert len(args) == 1, "typing.List takes just one arg"
            annotation = args[0]
            self._is_list = True

        self._is_dataframe = False
        if annotation is not None and issubclass(annotation, DataFrame):
            self._is_dataframe = True
            # For features, annotations like DataFrame[User.id] are not allowed
            # Annotations like these are only allowed in resolvers
            # So, error here.
            # if annotation.references_feature_set is None:
            #     raise TypeError("DF has no underlying type")
            annotation = annotation.references_feature_set

        self._is_scalar = annotation is not None and not issubclass(annotation, Features)

        if self._is_list and not self._is_scalar:
            raise TypeError(
                f"Invalid type annotation for feature{error_ctx}: Lists must be of scalar types, not {self._parsed_annotation}"
            )
        if self._is_dataframe and self._is_scalar:
            raise TypeError(
                f"Invalid type annotation for feature{error_ctx}: Dataframes must be of Features types, not {self._parsed_annotation}"
            )

        self._underlying = annotation

    @property
    def is_nullable(self) -> bool:
        """Whether the type annotation is nullable."""
        if self._parsed_annotation is None:
            self._parse_annotation()
        assert self._is_nullable is not None
        return self._is_nullable

    @property
    def underlying(self) -> type:
        """The underlying type annotation from the annotation."""
        if self.parsed_annotation is None:
            self._parse_annotation()
        if self._underlying is None:
            raise TypeError("There is no underlying type")
        return self._underlying

    @underlying.setter
    def underlying(self, underlying: Optional[type]):
        self._underlying = underlying

    @property
    def is_dataframe(self) -> bool:
        """Whether the type annotation is a dataframe."""
        if self._parsed_annotation is None:
            self._parse_annotation()
        assert self._is_dataframe is not None
        return self._is_dataframe

    @property
    def is_list(self) -> bool:
        """Whether the type annotation is a list."""
        if self._parsed_annotation is None:
            self._parse_annotation()
        assert self._is_list is not None
        return self._is_list

    @property
    def is_scalar(self) -> bool:
        """Whether the type annotation is a scalar type (i.e. not a Features type)."""
        if self._parsed_annotation is None:
            self._parse_annotation()
        assert self._is_scalar is not None
        return self._is_scalar


@dataclasses.dataclass
class HasOnePathObj:
    parent: Feature
    child: Feature
    parent_to_child_attribute_name: str


class FeatureWrapper:
    """The FeatureWrapper emulates DataFrames and nested has-one relationships when used as a type annotation or within a filter."""

    def __init__(self, feature: Feature) -> None:
        # Binding as a private variable as not to have naming conflicts user's use of __getattr__
        self._chalk_feature = feature

    def __hash__(self):
        return id(self)

    def __gt__(self, other):
        return Filter(self, ">", other)

    def __ge__(self, other):
        return Filter(self, ">=", other)

    def __lt__(self, other):
        return Filter(self, "<", other)

    def __le__(self, other):
        return Filter(self, "<=", other)

    def __ne__(self, other):
        return Filter(self, "!=", other)

    def __eq__(self, other):
        return Filter(self, "==", other)

    def __and__(self, other):
        return Filter(self, "and", other)

    def __or__(self, other):
        return Filter(self, "or", other)

    def __repr__(self):
        return f"FeatureWrapper(fqn={self._chalk_feature.namespace}.{self._chalk_feature.name}, typ={self._chalk_feature.typ})"

    def __str__(self):
        return str(self._chalk_feature)

    def in_(self, examples):
        return Filter(self._chalk_feature, "in", examples)

    def __getitem__(self, item):
        from chalk.df.ChalkASTParser import ChalkASTParser
        from chalk.features.dataframe import DataFrame

        if self._chalk_feature.typ is not None and issubclass(self._chalk_feature.typ.parsed_annotation, DataFrame):
            f_copy = FeatureWrapper(copy.copy(self._chalk_feature))

            with ChalkASTParser.ignore_ast_parse_errors():
                f_copy._chalk_feature.typ = FeatureTypeAnnotation(
                    underlying=self._chalk_feature.typ.parsed_annotation[item]
                )

            return f_copy
        raise TypeError(f"Feature {self} is not subscriptable.")

    def __getattr__(self, item: str):
        # Passing through __getattr__ on has_one features, as users can use getattr
        # notation in annotations for resolvers
        if item.startswith("__"):
            # Short-circuiting on the dunders to be compatible with copy.copy
            raise AttributeError(item)
        if self._chalk_feature.typ is not None and issubclass(self._chalk_feature.typ.underlying, Features):
            for f in self._chalk_feature.typ.underlying.features:
                assert isinstance(f, Feature), f"HasOne feature {f} does not inherit from FeaturesBase"
                if f.attribute_name == item:
                    return FeatureWrapper(self._chalk_feature.copy_with_path(f))
        raise AttributeError(item)


def unwrap_feature(maybe_feature_wrapper: Any) -> Feature:
    """Unwrap a class-annotated FeatureWrapper instance into the underlying feature.

    For example:

    .. code-block::

        @features
        class FooBar:
            foo: str
            bar: int

        type(FooBar.foo) is FeatureWrapper
        type(unwrap_feature(FooBar.foo)) is Feature
    """
    if isinstance(maybe_feature_wrapper, FeatureWrapper):
        return maybe_feature_wrapper._chalk_feature
    raise TypeError(f"{maybe_feature_wrapper} is of type {type(maybe_feature_wrapper)}, expecting type FeatureWrapper")


class Feature:
    def __init__(
        self,
        name: Optional[str] = None,
        attribute_name: Optional[str] = None,
        namespace: Optional[str] = None,
        features_cls: Optional[Type[Features]] = None,
        typ: Optional[Union[FeatureTypeAnnotation, type]] = None,
        version: Optional[int] = None,
        description: Optional[str] = None,
        owner: Optional[str] = None,
        tags: Optional[Tags] = None,
        primary: bool = False,
        max_staleness: Optional[Duration] = None,
        etl_offline_to_online: bool = False,
        encoder: Optional[Callable[[Any], JsonValue]] = None,
        decoder: Optional[Callable[[JsonValue], Any]] = None,
        join: Optional[Union[Callable[[], Filter], Filter]] = None,
        path: Tuple[HasOnePathObj, ...] = (),
        is_feature_time: bool = False,
    ):
        self.typ = FeatureTypeAnnotation(underlying=typ) if isinstance(typ, type) else typ
        self.features_cls = features_cls
        self.name = name
        self.attribute_name = attribute_name  # the attribute name for the feature in the @features class (in case if the name is specified differently)
        self.namespace = namespace
        self.path = path
        self.version = version
        self.description = description
        self.owner = owner
        self.tags = tags
        self.primary = primary
        self.max_staleness = max_staleness
        self.etl_offline_to_online = etl_offline_to_online
        self.encoder = encoder
        self.decoder = decoder
        self.is_feature_time = is_feature_time
        self._join = join

    def __str__(self):
        if len(self.path) > 0:
            return ".".join(
                [self.path[0].parent.namespace or "<unknown>"]
                + [(x.parent.name or "<unknown>") for x in self.path or []]
                + [self.name or "<unknown>"]
            )
        return f"{self.namespace}.{self.name}"

    def __repr__(self):
        return f"Feature(fqn={self.namespace}.{self.name}, typ={self.typ})"

    @property
    def fqn(self) -> str:
        return ".".join((x for x in [self.namespace, self.name] if isinstance(x, str)))

    @property
    def is_has_one(self):
        # A feature is a has-one relationship if the type is another singleton features cls and there is a join condition
        assert self.typ is not None
        # Need to short-circuit if it is a dataframe, as datafames might not have an underlying
        return not self.typ.is_dataframe and self.join is not None

    @property
    def is_has_many(self):
        assert self.typ is not None
        return self.typ.is_dataframe and self.join is not None

    @property
    def is_list(self):
        assert self.typ is not None
        return self.typ.is_list

    @property
    def has_resolved_join(self):
        return self._join is not None

    @property
    def join(self) -> Optional[Filter]:
        if self._join is not None:
            # Join was explicitly specified
            return self._join() if callable(self._join) else self._join
        # Attempt to extract the join condition from the foreign feature
        assert self.typ is not None
        foreign_features = self.typ.underlying
        if not issubclass(foreign_features, Features):
            return None
        assert self.features_cls is not None
        joins: List[Tuple[str, Filter]] = []  # Tuple of (name, Join)
        for f in foreign_features.features:
            assert isinstance(f, Feature)
            assert f.typ is not None
            if f.typ.underlying is self.features_cls and f.has_resolved_join:
                assert f.join is not None
                assert f.name is not None
                join = f.join() if callable(f.join) else f.join
                joins.append((f.name, join))
        if len(joins) == 0:
            # It's a nested feature
            return None
        # TODO(Ravi): Enable this check. But let's see if we can be smarter about which join to automatically use, if there are multiple
        # if len(joins) > 1:
        #     assert self.features_cls is not None
        #     raise ValueError(
        #         f"Multiple join conditions exist for {self.features_cls.__name__} and {foreign_features.__name__} on keys: "
        #         + f", ".join(f'{foreign_features.__name__}.{name}' for (name, _) in joins)
        #     )
        join = joins[0][1]
        if callable(join):
            join = join()
        self._join = join
        return join

    def copy_with_path(
        self,
        child: Feature,
    ):
        child_copy = copy.copy(child)
        assert child.attribute_name is not None
        child_copy.path = tuple(
            (
                *self.path,
                HasOnePathObj(
                    parent=self,
                    child=child,
                    parent_to_child_attribute_name=child.attribute_name,
                ),
            )
        )
        return child_copy


class Filter:
    def __init__(
        self,
        feature: Any,  # think of this as "LHS"
        operation: str,
        other: Any,  # think of this as "RHS", a feature or a constant
    ) -> None:
        # Feature or other could be another feature, filter, featuretime, literal
        # Other could also be a sequence (in the case of operation = "in")
        self.operation = operation
        if isinstance(feature, FeatureWrapper):
            feature = unwrap_feature(feature)
        self.feature = feature
        if self.operation == "in":
            if not isinstance(other, collections.abc.Iterable):
                raise ValueError("The RHS must be an iterable for operation='in'")
            other = {unwrap_feature(x) if isinstance(x, FeatureWrapper) else x for x in other}
        else:
            if isinstance(other, FeatureWrapper):
                other = unwrap_feature(other)
        self.other = other

    def __hash__(self) -> int:
        return hash((self.feature, self.operation, self.other))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Filter):
            return NotImplemented
        return self.feature == other.feature and self.operation == other.operation and self.other == other.other

    def __and__(self, other):
        return Filter(self, "and", other)

    def __or__(self, other):
        return Filter(self, "or", other)

    def __bool__(self):
        # If the operation is == or !=, then compare the underlying filters or features
        # If x and y are features, then (x == y) is overloaded to return a Filter
        # This allows bool(x == y) to return True iff the x is y, while (x == y) is still a Filter instance
        # Using identity instead of equality to prevent an infinite loop
        if self.operation == "==":
            return self.feature is self.other
        if self.operation == "!=":
            return self.feature is not self.other
        # Non-equality comparisons must be evaluated manually -- e.g. if x and y are Features
        # (x < y) would return a Filter, but bool(x < y) doesn't really make sense. So, raise.
        return NotImplemented

    def referenced_features(self) -> Set[Feature]:
        from chalk.parsed.feature import FeatureType

        result = set()
        if isinstance(self.feature, Feature):
            result.add(self.feature)

        if isinstance(self.other, Feature):
            result.add(self.other)
        if isinstance(self.feature, FeatureType):
            pass
        if isinstance(self.other, FeatureType):
            pass

        return result

    def __repr__(self):
        return f"Filter({self.feature} {self.operation} {self.other})"


class FeatureSetBase:
    """Registry containing all @features classes."""

    registry: Dict[str, Type[Features]] = {}  # mapping of fqn to Features cls

    def __init__(self) -> None:
        raise RuntimeError("FeatureSetBase should never be instantiated")


class FeaturesMeta(type):
    """Metaclass for classes decorated with ``@features``.

    This metaclass allows for:

    1.  Classes annotated with @features to pass the
        ``isinstance(x, Features)`` and ``issubclass(X, Features)`` checks.
    2.  ``Features[Feature1, Feature2]`` annotations to return subclasses of Features, so they can be used as proper type annotations.
    """

    def __subclasscheck__(self, subcls: type) -> bool:
        if not isinstance(subcls, type):
            raise TypeError(f"{subcls} is not a type")
        if hasattr(subcls, "__is_features__"):
            return getattr(subcls, "__is_features__")
        return False

    def __instancecheck__(self, instance: object) -> bool:
        return self.__subclasscheck__(type(instance))

    # This lets us subscript by the features
    def __getitem__(cls, item: Any) -> Type[Features]:
        from chalk.features.dataframe import DataFrame
        from chalk.parsed.feature import FeatureType

        # Typing the `__getitem__` as any, since the @features members are typed as the underlying data structure
        # But, they should always be features or a tuple of features
        if isinstance(item, type) and issubclass(item, Features):
            item = [f for f in item.features if isinstance(f, (Feature, FeatureWrapper))]
        item = ensure_tuple(item)
        item = tuple(x._chalk_feature if isinstance(x, FeatureWrapper) else x for x in item)
        for x in item:
            if isinstance(x, str):
                raise TypeError(
                    f'String features like {cls.__name__}["{x}"] are unsupported. Instead, replace with "{cls.__name__}[{x}]"'
                )
            if isinstance(x, (Feature, FeatureType)) or isinstance(x, type) and issubclass(x, DataFrame):
                continue

            raise TypeError(f"Invalid feature type: {x}")
        cls = cast(Type[Features], cls)

        new_features = tuple([*cls.features, *item])
        assert len(new_features) > 0

        class SubFeatures(cls):
            features = new_features

        return SubFeatures

    def __repr__(cls) -> str:
        cls = cast(Type[Features], cls)
        return f"Features[{', '.join(str(f) for f in cls.features)}]"

    @property
    def namespace(cls):
        cls = cast(Type[Features], cls)
        namespaces = [x.namespace for x in cls.features]
        return get_unique_item(namespaces, name=f"{cls.__name__} feature namespaces")


class Features(metaclass=FeaturesMeta):
    """Features base class.

    This class is rarely instantiated or directly inherited. However, classes
    annotated with @features can be thought of as inheriting from this class.
    It can be used with `isinstance()` and ``issubclass`` checks, as well as for
    typing.
    """

    # Internally, the Features class is instantiated when results come through, and
    # results are bound to instances of this class via attributes
    __chalk_primary__: ClassVar[Optional[Feature]]  # The primary key feature
    __chalk_ts__: ClassVar[Optional[Feature]]  # The timestamp feature
    features: Tuple[Union[Feature, FeatureType, Type[DataFrame]], ...] = ()
    __is_features__: bool = True

    # When constructing results, this class is instantiated


@dataclasses.dataclass
class TimeDelta:
    weeks_ago: int = 0
    days_ago: int = 0
    hours_ago: int = 0
    minutes_ago: int = 0
    seconds_ago: int = 0
    milliseconds_ago: int = 0
    microseconds_ago: int = 0

    def to_std(self) -> timedelta:
        # Returning the negative inverse since the feature is defined in the past (e.g. days **ago**)
        return -timedelta(
            weeks=self.weeks_ago,
            days=self.days_ago,
            hours=self.hours_ago,
            minutes=self.minutes_ago,
            seconds=self.seconds_ago,
            milliseconds=self.milliseconds_ago,
            microseconds=self.microseconds_ago,
        )


CHALK_TS_FEATURE = Feature(name="CHALK_TS", namespace="__chalk__", typ=datetime)
ChalkTime = CHALK_TS_FEATURE


def before(
    weeks_ago: int = 0,
    days_ago: int = 0,
    hours_ago: int = 0,
    minutes_ago: int = 0,
    seconds_ago: int = 0,
    milliseconds_ago: int = 0,
    microseconds_ago: int = 0,
) -> Any:
    return Filter(
        feature=CHALK_TS_FEATURE,
        operation="<=",
        other=TimeDelta(
            weeks_ago=weeks_ago,
            days_ago=days_ago,
            hours_ago=hours_ago,
            minutes_ago=minutes_ago,
            seconds_ago=seconds_ago,
            milliseconds_ago=milliseconds_ago,
            microseconds_ago=microseconds_ago,
        ),
    )


def after(
    weeks_ago: int = 0,
    days_ago: int = 0,
    hours_ago: int = 0,
    minutes_ago: int = 0,
    seconds_ago: int = 0,
    milliseconds_ago: int = 0,
    microseconds_ago: int = 0,
) -> Any:
    return Filter(
        feature=CHALK_TS_FEATURE,
        operation=">=",
        other=TimeDelta(
            weeks_ago=weeks_ago,
            days_ago=days_ago,
            hours_ago=hours_ago,
            minutes_ago=minutes_ago,
            seconds_ago=seconds_ago,
            milliseconds_ago=milliseconds_ago,
            microseconds_ago=microseconds_ago,
        ),
    )


UNPARSED_TIMESTAMP = Feature(typ=datetime)



import json
import logging
import os
import sys
import traceback
from pathlib import Path

import click
from pydantic import ValidationError

from chalk import importer
from chalk.config.project_config import ProjectSettings, load_project_config
from chalk.parsed.user_types_to_json import get_registered_types_as_json


@click.group()
@click.option(
    "--log-level",
    default="info",
    help="The logging level. One of DEBUG, INFO, WARNING, ERROR, and CRITICAL (case-insensitive).",
)
def cli(log_level: str):
    level = getattr(logging, log_level.upper())
    logging.basicConfig(
        format="%(asctime)s %(levelname)s: %(message)s",
        datefmt="%m/%d/%Y %I:%M:%S %p",
        level=level,
    )


@cli.command("list")
@click.argument(
    "directory",
    default=None,
    required=False,
)
def list_cmd(directory):
    scope_to = Path(directory or os.getcwd())
    try:
        failed = importer.import_all_python_files()
        print(get_registered_types_as_json(scope_to, failed))
    except Exception:
        ex_type, ex_value, ex_traceback = sys.exc_info()
        relevant_traceback = f"""{os.linesep.join(traceback.format_tb(ex_traceback))}
\n{ex_type.__name__}: {str(ex_value)}
"""
        print(json.dumps(dict(failed=[dict(traceback=relevant_traceback)]), indent=2))


@cli.command("config")
def config():
    json_response: str
    try:
        model = load_project_config() or ProjectSettings()
        json_response = model.json()
    except ValidationError as e:
        json_response = json.dumps({"error": str(e)})

    print(json_response)


if __name__ == "__main__":
    cli()



import enum
import unittest
from datetime import date, datetime

import pendulum

from chalk.features import Feature, feature, features
from chalk.features.feature import unwrap_feature
from chalk.serialization.codec import FeatureCodec


class CustomClass:
    def __init__(self, w: str):
        self.w = w

    def __eq__(self, other):
        return isinstance(other, CustomClass) and self.w == other.w


class Gender(enum.Enum):
    M = "m"
    F = "f"
    X = "x"


@features
class Hello:
    a: str
    b: int
    c: datetime
    d: Gender
    e: date
    fancy: CustomClass = feature(encoder=lambda x: x.w, decoder=lambda x: CustomClass(x))


class FeatureCodecTestCase(unittest.TestCase):
    codec = FeatureCodec()

    def _check_roundtrip(self, f: Feature, value):
        encoded = self.codec.encode(f, value)
        decoded = self.codec.decode(f, encoded)
        self.assertTrue(decoded == value)

    def test_datetime(self):
        serialized = "2022-04-08T22:26:03.303000+00:00"
        decoded = self.codec.decode(unwrap_feature(Hello.c), serialized)
        self.assertEqual(decoded, pendulum.parse(serialized))
        re_encoded = self.codec.encode(unwrap_feature(Hello.c), decoded)
        self.assertEqual(serialized, re_encoded)

    def test_custom_codecs(self):
        self._check_roundtrip(unwrap_feature(Hello.fancy), CustomClass("hihi"))

    def test_gender(self):
        self._check_roundtrip(unwrap_feature(Hello.d), Gender.F)

    def test_date(self):
        self.assertEqual(
            "2022-04-08",
            self.codec.encode(unwrap_feature(Hello.e), date.fromisoformat("2022-04-08")),
        )
        self._check_roundtrip(unwrap_feature(Hello.e), date.fromisoformat("2022-04-08"))



import inspect
import weakref
from datetime import datetime
from typing import Any, List, Optional, Type, TypeVar, Union, overload

from chalk.features import DataFrame, Feature, Filter
from chalk.features.feature import Features, FeatureSetBase, FeatureWrapper, HasOnePathObj, unwrap_feature
from chalk.features.resolver import Cron, OfflineResolver, OnlineResolver, SinkResolver
from chalk.features.tag import Tags
from chalk.parsed.feature import (
    DataFrameType,
    FeatureSetType,
    FeatureTimeFeatureType,
    FeatureType,
    FilterParsed,
    HasManyFeatureType,
    HasOneFeatureType,
    HasOnePathObjParsed,
    InputFeatureType,
    ScalarFeatureType,
)
from chalk.parsed.resolver import (
    CronFilterWithFeatureArgs,
    OfflineResolverParsed,
    OnlineResolverParsed,
    SinkResolverParsed,
)
from chalk.utils.duration import ScheduleOptions

T = TypeVar("T")
TFeatureType = TypeVar("TFeatureType", bound=FeatureType)


class APIToProcessorConverter:
    def __init__(self):
        self._cache = weakref.WeakKeyDictionary()

    def _convert_filter(self, f: Union[Filter, FilterParsed]) -> FilterParsed:
        if isinstance(f, FilterParsed):
            return f

        def convert_if_feature(v: Any):
            if isinstance(v, Feature):
                return self._convert_scalar_feature(v)
            elif isinstance(v, Filter):
                return self._convert_filter(v)
            return v

        return FilterParsed(
            lhs=convert_if_feature(f.feature),
            operation=f.operation,
            rhs=convert_if_feature(f.other),
        )

    def _convert_scalar_feature(self, f: Feature) -> ScalarFeatureType:
        cached_val = self._cache.get(f)
        if cached_val is not None:
            assert isinstance(cached_val, ScalarFeatureType)
            return cached_val

        if f.name is None or f.namespace is None:
            raise ValueError(f"Referenced feature is missing namespace or name {f.namespace}.{f.name}.")

        if f.typ is None:
            raise ValueError(f"Feature missing type {f.namespace}.{f.name}")

        converted = ScalarFeatureType(
            name=f.name,
            namespace=f.namespace,
            type=f.typ.underlying,
            is_nullable=f.typ.is_nullable,
            is_list=f.typ.is_list,
            primary=f.primary,
            version=f.version,
            description=f.description,
            owner=f.owner,
            tags=f.tags,
            max_staleness=f.max_staleness,
            etl_offline_to_online=f.etl_offline_to_online,
            encoder=f.encoder,
            decoder=f.decoder,
        )
        self._cache[f] = converted
        return converted

    @staticmethod
    def _convert_feature_time(t: Feature) -> FeatureTimeFeatureType:
        assert t.is_feature_time
        assert t.typ is not None
        assert issubclass(t.typ.underlying, datetime)
        assert t.name is not None
        assert t.namespace is not None
        return FeatureTimeFeatureType(t.name, t.namespace)

    def _convert_has_one(self, f: Feature) -> HasOneFeatureType:
        assert f.is_has_one
        if self._cache.get(f) is not None:
            val = self._cache.get(f)
            assert isinstance(val, HasOneFeatureType)
            return val

        if f.name is None or f.namespace is None:
            raise ValueError(f"Feature wasn't named {f.namespace}.{f.name}")

        if f.typ is None:
            raise ValueError(f"Feature missing typ {f.namespace}.{f.name}")

        if f.join is None:
            raise ValueError(f"Feature missing join {f.namespace}.{f.name}")

        assert issubclass(f.typ.underlying, Features), "has-one relationships must reference a Features cls"
        ret = HasOneFeatureType(
            namespace=f.namespace,
            name=f.name,
            join=self._convert_filter(f.join),
            is_nullable=f.typ.is_nullable,
        )
        self._cache[f] = ret
        ret.type = self.convert_type(f.typ.underlying)
        return ret

    def _convert_has_many(self, f: Feature) -> HasManyFeatureType:
        assert f.is_has_many
        cached_val = self._cache.get(f)
        if cached_val is not None:
            assert isinstance(cached_val, HasManyFeatureType)
            return cached_val

        if f.name is None or f.namespace is None:
            raise ValueError(f"Feature wasn't named: {f.namespace}.{f.name}")

        assert f.join is not None, "has many feature type should have join"
        assert f.typ is not None, "type should be defined"

        ret = HasManyFeatureType(
            namespace=f.namespace,
            name=f.name,
            join=self._convert_filter(f.join),
            type=None,
        )
        self._cache[f] = ret
        parsed_annotation = f.typ.parsed_annotation
        assert issubclass(parsed_annotation, DataFrame)
        converted_type = self.convert_type(parsed_annotation)
        assert isinstance(converted_type, DataFrameType)
        ret.type = converted_type
        return ret

    def _convert_environment(self, e: Optional[Union[List[str], str]]) -> Optional[List[str]]:
        return [e] if isinstance(e, str) else e

    def _convert_tags(self, tags: Optional[Tags]) -> Optional[List[str]]:
        if tags is None:
            return None
        if isinstance(tags, list):
            return tags
        if isinstance(tags, str):
            return [tags]
        raise ValueError("Tags should be a list of strings or a string or None")

    def _convert_path(self, o: Feature) -> List[HasOnePathObjParsed]:
        return [self._convert_has_one_path_obj(p) for p in o.path]

    def _wrap_as_input_ft(self, converted: TFeatureType, raw: Feature) -> Union[InputFeatureType, TFeatureType]:
        """"""
        if len(raw.path) < 1:
            return converted

        i = InputFeatureType(underlying=converted, path=self._convert_path(raw))
        return i

    def feature_from_string(self, s: str) -> Optional[Feature]:
        split = s.split(".")
        current = None
        if len(split) > 0:
            base = FeatureSetBase.registry.get(split[0])
            current = base
            for p in split[1:]:
                if current is not None:
                    try:
                        current = getattr(current, p)
                    except AttributeError:
                        current = None
        if isinstance(current, FeatureWrapper):
            current = unwrap_feature(current)
        assert current is None or isinstance(current, Feature)
        return current

    def convert_target(self, raw: Union[Feature, FeatureWrapper]) -> FeatureType:
        if isinstance(raw, FeatureWrapper):
            raw = unwrap_feature(raw)
        converted = self.convert_type(raw)
        return self._wrap_as_input_ft(converted, raw)

    def _convert_has_one_path_obj(self, t: HasOnePathObj) -> HasOnePathObjParsed:
        parent = self.convert_type(t.parent)
        assert isinstance(parent, HasOneFeatureType)
        return HasOnePathObjParsed(
            parent=parent,
            child=self.convert_type(t.child),
            parent_to_child_attribute_name=t.parent_to_child_attribute_name,
        )

    def _convert_sink_resolver(self, r: SinkResolver) -> SinkResolverParsed:
        return SinkResolverParsed(
            function_definition=r.function_definition,
            filename=r.filename,
            fqn=r.fqn,
            doc=r.doc,
            inputs=[self._wrap_as_input_ft(self.convert_type(i), i) for i in r.inputs],
            raw_fn=r.fn,
            environment=self._convert_environment(r.environment),
            tags=self._convert_tags(r.tags),
            machine_type=r.machine_type,
            buffer_size=r.buffer_size,
            debounce=r.debounce,
            max_delay=r.max_delay,
            upsert=r.upsert,
        )

    def _convert_cron_filter(self, cron: Optional[Union[ScheduleOptions, Cron]]) -> Optional[CronFilterWithFeatureArgs]:
        if cron is not None and isinstance(cron, Cron) and cron.filter is not None:
            sig = inspect.signature(cron.filter)
            filter_features = [FeatureType.of(parameter.annotation) for parameter in sig.parameters.values()]
            return CronFilterWithFeatureArgs(
                filter=cron.filter,
                feature_args=filter_features,
            )
        else:
            return None

    def _convert_offline_resolver(self, r: OfflineResolver) -> OfflineResolverParsed:
        output = self.convert_type(r.output)
        return OfflineResolverParsed(
            function_definition=r.function_definition,
            filename=r.filename,
            fqn=r.fqn,
            doc=r.doc,
            inputs=[self._wrap_as_input_ft(self.convert_type(i), i) for i in r.inputs],
            output=output,
            raw_fn=r.fn,
            fn=r.fn,
            environment=self._convert_environment(r.environment),
            tags=self._convert_tags(r.tags),
            cron=r.cron,
            cron_filter=self._convert_cron_filter(r.cron),
            machine_type=r.machine_type,
        )

    def _convert_online_resolver(self, r: OnlineResolver) -> OnlineResolverParsed:
        output = self.convert_type(r.output)
        return OnlineResolverParsed(
            function_definition=r.function_definition,
            filename=r.filename,
            fqn=r.fqn,
            doc=r.doc,
            inputs=[self._wrap_as_input_ft(self.convert_type(i), i) for i in r.inputs],
            output=output,
            raw_fn=r.fn,
            fn=r.fn,
            environment=self._convert_environment(r.environment),
            tags=self._convert_tags(r.tags),
            cron=r.cron,
            cron_filter=self._convert_cron_filter(r.cron),
            machine_type=r.machine_type,
            when=self._convert_filter(r.when) if r.when is not None else None,
        )

    @overload
    def convert_type(
        self,
        t: Union[Feature, FeatureWrapper],
    ) -> Union[HasOneFeatureType, HasManyFeatureType, FeatureTimeFeatureType, ScalarFeatureType]:
        ...

    @overload
    def convert_type(
        self,
        t: FeatureType[T],
        parent: Optional[Feature] = None,
    ) -> FeatureType[T]:
        ...

    @overload
    def convert_type(
        self,
        t: Type[DataFrame],
    ) -> DataFrameType:
        ...

    @overload
    def convert_type(
        self,
        t: Type[Features],
    ) -> FeatureSetType:
        ...

    @overload
    def convert_type(self, t: Union[OnlineResolver, OnlineResolverParsed]) -> OnlineResolverParsed:
        ...

    @overload
    def convert_type(self, t: Union[OfflineResolver, OfflineResolverParsed]) -> OfflineResolverParsed:
        ...

    @overload
    def convert_type(self, t: Union[SinkResolver, SinkResolverParsed]) -> SinkResolverParsed:
        ...

    @overload
    def convert_type(self, t: Union[HasOnePathObj, HasOnePathObjParsed]) -> HasOnePathObjParsed:
        ...

    def convert_type(
        self,
        t,
    ):
        if isinstance(t, type):
            if issubclass(t, DataFrame):
                cols = [self._wrap_as_input_ft(self.convert_type(f), f) for f in t.columns]
                column_filters = [self._convert_filter(f) for f in t.filters]
                return DataFrameType(columns=cols, filters=column_filters)

            if issubclass(t, Features):
                return FeatureSetType[Any](
                    features=[self.convert_type(f) for f in t.features],
                )

        if isinstance(t, FeatureWrapper):
            t = unwrap_feature(t)

        if isinstance(t, Feature):
            assert t.typ is not None

            if t.is_has_one:
                return self._convert_has_one(t)

            elif t.is_has_many:
                return self._convert_has_many(t)

            elif t.is_feature_time:
                return self._convert_feature_time(t)

            else:
                return self._convert_scalar_feature(t)

        if isinstance(t, OnlineResolver):
            return self._convert_online_resolver(t)

        if isinstance(t, OfflineResolver):
            return self._convert_offline_resolver(t)

        if isinstance(t, HasOnePathObj):
            return self._convert_has_one_path_obj(t)

        if isinstance(t, SinkResolver):
            return self._convert_sink_resolver(t)

        if isinstance(
            t,
            (
                FeatureType,
                HasOnePathObjParsed,
                DataFrameType,
                FeatureSetType,
                OnlineResolverParsed,
                OfflineResolverParsed,
                SinkResolverParsed,
            ),
        ):
            return t

        raise TypeError(f"Unexpected type: {t}")


DEFAULT_CONVERTER = APIToProcessorConverter()



from sqlalchemy.engine.url import URL

from chalk.sql.base.sql_source import BaseSQLSource


class BigQuerySource(BaseSQLSource):
    def local_engine_url(self) -> URL:
        raise NotImplemented("Local BigQuery client not implemented")



import ast
import contextlib
import itertools
from typing import Callable, List, Optional

from pure_eval.core import Evaluator
from varname import VarnameRetrievingError
from varname.ignore import IgnoreList
from varname.utils import get_node_by_frame

from chalk.features import Filter
from chalk.utils.collections import ensure_tuple


class ChalkASTParser:
    _ignore_ast_parse_errors: bool = False

    def __init__(self, condition: Callable[[ast.AST], bool]):
        self.func_frame, self.func_node = self._get_func_frame_and_nodes(condition)
        self.evaluator = Evaluator.from_frame(self.func_frame)

    def _get_func_frame_and_nodes(self, condition: Callable[[ast.AST], bool]):
        # Start up three frames -- this function, ChalkASTParser.__init__, and the calling function
        ignore_list = IgnoreList.create(
            None,
            ignore_lambda=False,
            ignore_varname=False,
        )
        for frame_num in itertools.count(start=3):
            frame = ignore_list.get_frame(frame_num)
            func_node = get_node_by_frame(frame)
            if condition(func_node):
                # This is the correct getitem frame.
                # It is important that the "slice" isn't an ast.Name, as otherwise it would be impossible to parse the expression
                # ast.Name would imply something like this:
                # def __getitem__(self, item):
                #    return self.df[item]  # <--- item is of type ast.name. We need to go one frame higher!
                return frame, func_node
        assert False, "unreachable"

    @classmethod
    @contextlib.contextmanager
    def ignore_ast_parse_errors(cls, ignore: bool = True):
        existing = cls._ignore_ast_parse_errors
        cls._ignore_ast_parse_errors = ignore
        try:
            yield
        finally:
            cls._ignore_ast_parse_errors = existing

    @classmethod
    def parse_dataframe_getitem(cls, item) -> List:
        item = ensure_tuple(item)

        parsed_items = []
        ast_indicies = []
        for idx, queried in enumerate(item):
            if isinstance(queried, (Filter, bool)):
                ast_indicies.append(idx)

            else:
                parsed_items.append(queried)
        if len(ast_indicies) > 0:
            try:
                parser = cls(lambda node: isinstance(node, ast.Subscript) and not isinstance(node.slice, ast.Name))
            except VarnameRetrievingError as e:
                if cls._ignore_ast_parse_errors:
                    parsed_items.extend((item[i] for i in ast_indicies))
                else:
                    raise ValueError(
                        "Could not parse DataFrame item. This may happen if you're using some other AST magic at the same time, such as pytest, ipython, macropy, or birdseye. "
                        "Try declaring you DataFrame operation on a new line."
                    ) from e
            else:
                func_node = parser.func_node
                assert isinstance(func_node, ast.Subscript)
                nodes = [func_node.slice]
                if isinstance(func_node.slice, ast.Tuple):
                    nodes = func_node.slice.elts
                items = [parser._parse_node(n) for idx, n in enumerate(nodes) if idx in ast_indicies]
                for idx in ast_indicies:
                    if idx >= len(ast_indicies):
                        # Pass through anything that wasn't parsed
                        items.append(item[idx])
                parsed_items.extend(items)

        return parsed_items

    @classmethod
    def parse_when(cls) -> Optional[Filter]:
        try:
            parser = cls(lambda node: isinstance(node, ast.Call))
            func_node = parser.func_node
            when_filter = None
            assert isinstance(func_node, ast.Call)
            when = next((k for k in func_node.keywords if k.arg == "when"), None)
            when_filter = parser._parse_node(when.value) if when else None
            return when_filter
        except VarnameRetrievingError:
            return None

    def _parse_node(self, n):
        from chalk.features.feature import Filter

        if isinstance(n, ast.Name):
            return self.evaluator[n]

        if isinstance(n, ast.Attribute):
            parent = self._parse_node(n.value)
            return getattr(parent, n.attr)

        if isinstance(n, ast.UnaryOp) and isinstance(n.op, ast.Not):
            expr = self._parse_node(n.operand)
            return Filter(expr, "not", None)

        if isinstance(n, ast.BoolOp):
            lhs = self._parse_node(n.values[0])
            rhs = self._parse_node(n.values[1])

            if isinstance(n.op, ast.And):
                return Filter(lhs, "and", rhs)

            if isinstance(n.op, ast.Or):
                return Filter(lhs, "or", rhs)

        if isinstance(n, ast.Compare):
            left = self._parse_node(n.left)
            right = self._parse_node(n.comparators[0])
            ast_op = n.ops[0]
            op = None

            if isinstance(ast_op, ast.NotIn):
                op = "not in"
            elif isinstance(ast_op, ast.In):
                op = "in"
            elif isinstance(ast_op, ast.Gt):
                op = ">"
            elif isinstance(ast_op, ast.GtE):
                op = ">="
            elif isinstance(ast_op, ast.Lt):
                op = "<"
            elif isinstance(ast_op, ast.LtE):
                op = "<="
            elif isinstance(ast_op, (ast.Eq, ast.Is)):
                op = "=="
            elif isinstance(ast_op, (ast.NotEq, ast.IsNot)):
                op = "!="

            if op:
                return Filter(feature=left, operation=op, other=right)
            else:
                raise ValueError("Invalid filter for DataFrame", ast.unparse(n))

        if isinstance(n, ast.Call):
            return eval(ast.unparse(n), self.func_frame.f_globals, self.func_frame.f_locals)

        try:
            return self.evaluator[n]
        except Exception:
            return eval(ast.unparse(n), self.func_frame.f_globals, self.func_frame.f_locals)



from dataclasses import dataclass
from typing import List, Optional, Union

from chalk.features import Features


@dataclass
class StreamUpdate:
    online: Optional[Union[Features, List[Features]]] = None
    offline: Optional[Union[Features, List[Features]]] = None



import dataclasses
from types import NoneType
from typing import TypeVar, Union

import pydantic

from chalk.parsed.duplicate_input_gql import (
    KafkaConsumerConfigGQL,
    UpsertDataFrameGQL,
    UpsertFeatureGQL,
    UpsertFeatureIdGQL,
    UpsertFeatureReferenceGQL,
    UpsertFeatureTimeKindGQL,
    UpsertFilterGQL,
    UpsertHasManyKindGQL,
    UpsertHasOneKindGQL,
    UpsertReferencePathComponentGQL,
    UpsertResolverGQL,
    UpsertResolverOutputGQL,
    UpsertScalarKindGQL,
    UpsertSinkResolverGQL,
    UpsertStreamResolverGQL,
)
from chalk.parsed.feature import (
    DataFrameType,
    FeatureSetType,
    FeatureTimeFeatureType,
    FeatureType,
    FilterParsed,
    HasManyFeatureType,
    HasOneFeatureType,
    HasOnePathObjParsed,
    InputFeatureType,
    ScalarFeatureType,
)
from chalk.parsed.resolver import OfflineResolverParsed, OnlineResolverParsed, SinkResolverParsed
from chalk.streams.KafkaSource import KafkaConsumerConfig
from chalk.streams.stream import StreamResolver

T = TypeVar("T")


try:
    import attrs
except ImportError:
    attrs = None


def _get_feature_reference(f: FeatureType):
    return UpsertFeatureIdGQL(
        name=f.name,
        namespace=f.namespace,
        fqn=f.fqn,
    )


def _convert_filter_val(v: Union[T, ScalarFeatureType[T]]):
    if isinstance(v, ScalarFeatureType):
        return _get_feature_reference(v)
    raise ValueError("Haven't handled filters that aren't feature references")
    # return j("Scalar", dict(value=v))


def convert_type(t: T):
    if isinstance(t, KafkaConsumerConfig):
        return KafkaConsumerConfigGQL(
            broker=list(t.broker) if isinstance(t.broker, str) else t.broker,
            topic=list(t.topic) if isinstance(t.topic, str) else t.topic,
            ssl_keystore_location=t.ssl_keystore_location,
            client_id_prefix=t.client_id_prefix,
            group_id_prefix=t.group_id_prefix,
            topic_metadata_refresh_interval_ms=t.topic_metadata_refresh_interval_ms,
            security_protocol=t.security_protocol,
        )

    if isinstance(t, StreamResolver):
        return UpsertStreamResolverGQL(
            fqn=t.fqn,
            kind="stream",
            config=convert_type(t.source.consumer_config),
            function_definition=t.function_definition,
            environment=[t.environment] if isinstance(t.environment, str) else t.environment,
            doc=t.fn.__doc__,
        )

    if isinstance(t, HasManyFeatureType):
        return UpsertFeatureGQL(
            id=UpsertFeatureIdGQL(
                fqn=t.fqn,
                name=t.name,
                namespace=t.namespace,
            ),
            has_many_kind=UpsertHasManyKindGQL(
                columns=[convert_type(x).id for x in t.type.columns],
                filters=[convert_type(x).id for x in t.type.filters],
                join=convert_type(t.join),
            ),
            max_staleness=None,
            description=None,
            owner=None,
        )

    if isinstance(t, ScalarFeatureType):
        tags = [t.tags] if isinstance(t.tags, str) else t.tags
        assert isinstance(t.type, type)
        scalar_kind = t.type.__name__
        base_classes = [x.__name__ for x in type.mro(t.type)]

        # Attrs and Dataclasses don't technically have base classes
        # Pydantic calls their base class BaseModel which is way too generic for string comparsiion
        # For simplicity on the server-side validation, we'll come up with our own "base class" names
        if dataclasses.is_dataclass(t.type):
            base_classes.append("__dataclass__")
        if attrs is not None and isinstance(t.type, type) and attrs.has(t.type):
            base_classes.append("__attrs__")
        if issubclass(t.type, pydantic.BaseModel):
            base_classes.append("__pydantic__")

        return UpsertFeatureGQL(
            id=UpsertFeatureIdGQL(
                fqn=t.fqn,
                name=t.name,
                namespace=t.namespace,
            ),
            scalar_kind=UpsertScalarKindGQL(
                scalar_kind=scalar_kind,
                primary=t.primary,
                base_classes=base_classes,
                version=t.version,
                has_encoder_and_decoder=t.encoder is not None and t.decoder is not None,
            ),
            tags=tags,
            max_staleness=t.max_staleness,
            description=t.description,
            owner=t.owner,
            etl_offline_to_online=t.etl_offline_to_online,
        )

    if isinstance(t, HasOneFeatureType):
        return UpsertFeatureGQL(
            id=UpsertFeatureIdGQL(
                fqn=t.fqn,
                name=t.name,
                namespace=t.namespace,
            ),
            has_one_kind=UpsertHasOneKindGQL(join=convert_type(t.join)),
            max_staleness=None,
            description=None,
            owner=None,
        )

    if isinstance(t, FeatureTimeFeatureType):
        return UpsertFeatureGQL(
            id=UpsertFeatureIdGQL(
                fqn=t.fqn,
                name=t.name,
                namespace=t.namespace,
            ),
            feature_time_kind=UpsertFeatureTimeKindGQL(),
            max_staleness=None,
            description=None,
            owner=None,
        )

    if isinstance(t, FeatureSetType):
        return UpsertResolverOutputGQL(
            dataframes=[
                convert_type(f) for f in t.features if not isinstance(f, (FeatureTimeFeatureType, ScalarFeatureType))
            ],
            features=[
                _get_feature_reference(f)
                for f in t.features
                if isinstance(f, (FeatureTimeFeatureType, ScalarFeatureType))
            ],
        )

    if isinstance(t, DataFrameType):
        return UpsertDataFrameGQL(
            columns=[_get_feature_reference(f) for f in t.columns],
            filters=[convert_type(f) for f in t.filters],
        )

    if isinstance(t, FilterParsed):
        return UpsertFilterGQL(
            lhs=_convert_filter_val(t.lhs),
            op=t.operation,
            rhs=_convert_filter_val(t.rhs),
        )

    if isinstance(t, HasOnePathObjParsed):
        return UpsertReferencePathComponentGQL(
            parent=_get_feature_reference(t.parent),
            child=_get_feature_reference(t.child),
            parent_to_child_attribute_name=t.parent_to_child_attribute_name,
        )

    if isinstance(t, InputFeatureType):
        return UpsertFeatureReferenceGQL(
            underlying=_get_feature_reference(t.underlying),
            path=t.path and [convert_type(p) for p in t.path],
        )

    if isinstance(t, SinkResolverParsed):
        tags = [t.tags] if isinstance(t.tags, str) else t.tags

        resolver = UpsertSinkResolverGQL(
            fqn=t.fqn,
            function_definition=t.function_definition,
            inputs=[
                convert_type(i) if isinstance(i, InputFeatureType) else convert_type(InputFeatureType(i, list()))
                for i in t.inputs
            ],
            environment=t.environment,
            tags=tags,
            doc=t.doc,
            machine_type=t.machine_type,
            buffer_size=t.buffer_size,
            debounce=t.debounce,
            max_delay=t.max_delay,
            upsert=t.upsert,
        )
        return resolver

    if isinstance(t, (OfflineResolverParsed, OnlineResolverParsed)):
        assert isinstance(
            t.cron, (NoneType, (NoneType, str))
        ), f"Only supporting cron as a string right now for {t.fqn}"
        tags = [t.tags] if isinstance(t.tags, str) else t.tags

        resolver = UpsertResolverGQL(
            fqn=t.fqn,
            kind="offline" if isinstance(t, OfflineResolverParsed) else "online",
            function_definition=t.function_definition,
            inputs=[
                convert_type(i) if isinstance(i, InputFeatureType) else convert_type(InputFeatureType(i, list()))
                for i in t.inputs
            ],
            output=convert_type(t.output),
            environment=t.environment,
            tags=tags,
            doc=t.doc,
            cron=t.cron,
            machine_type=t.machine_type,
        )
        return resolver

    return t



from ..feature_n import Feature2
from .dataframe import DataFrame, DataFrameMeta
from .feature import ChalkTime, Feature, Features, FeatureSetBase, Filter, after, before
from .feature_set import feature, feature_time, features, has_many, has_one
from .hooks import after_all, before_all
from .resolver import Cron, MachineType, ScheduleOptions, offline, online, sink
from .tag import Environments, Tags

__all__ = [
    "Feature2",
    "before",
    "after",
    "Filter",
    "Feature",
    "Features",
    "FeatureSetBase",
    "ChalkTime",
    "features",
    "has_one",
    "feature",
    "has_many",
    "feature_time",
    "DataFrame",
    "DataFrameMeta",
    "Environments",
    "Tags",
    "online",
    "sink",
    "offline",
    "Cron",
    "ScheduleOptions",
    "MachineType",
    "before_all",
    "after_all",
]



import collections.abc
from typing import Any, Dict, Generator, List, Optional, Sequence, Set, TypeVar, Union

T = TypeVar("T")


def flatten(v: Sequence[Union[T, Sequence[Union[T, Sequence[T]]]]]) -> List[T]:
    ret = []
    for x in v:
        if isinstance(x, collections.abc.Sequence) and not isinstance(x, (str, bytes, bytearray)):
            ret.extend(flatten(x))
        else:
            ret.append(x)
    return ret


def chunks(lst: Sequence[T], n: int) -> Generator[Sequence[T], None, None]:
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def ensure_tuple(x: Union[T, Sequence[T], Dict[Any, T], None]):
    """Converts ``x`` into a tuple.
    * If ``x`` is ``None``, then ``tuple()`` is returned.
    * If ``x`` is a tuple, then ``x`` is returned as-is.
    * If ``x`` is a list, then ``tuple(x)`` is returned.
    * If ``x`` is a dict, then ``tuple(v for v in x.values())`` is returned.
    Otherwise, a single element tuple of ``(x,)`` is returned.
    Args:
        x (Any): The input to convert into a tuple.
    Returns:
        tuple: A tuple of ``x``.
    """
    # From https://github.com/mosaicml/composer/blob/020ca02e3848ee8fb6b7fff0c8123f597b05be8a/composer/utils/iter_helpers.py#L40
    if x is None:
        return ()
    if isinstance(x, (str, bytes, bytearray)):
        return (x,)
    if isinstance(x, collections.abc.Sequence):
        return tuple(x)
    if isinstance(x, dict):
        return tuple(x.values())
    return (x,)


def flatten_sets(v: List[Set[T]]) -> Set[T]:
    return set().union(*v)


def get_unique_item(collection: Sequence[Optional[T]], name: str) -> T:
    item = None
    for x in collection:
        if x is None:
            raise ValueError(f"Item in {name} is None")
        if item is not None:
            if x != item:
                raise ValueError(f"Multiple values in {name} are not permitted. Found {x}, {item}")
        item = x
    if item is None:
        raise ValueError(f"There should be at least one item in {name}")
    return item



from dataclasses import dataclass
from typing import Callable, List, Optional, Protocol, Sequence, TypeVar, Union

from chalk.features.resolver import Cron, FilterFunction, MachineType
from chalk.parsed.feature import FeatureSetType, FeatureType, FilterParsed, QueryInputType
from chalk.utils.collections import get_unique_item
from chalk.utils.duration import Duration, ScheduleOptions

T = TypeVar("T")


@dataclass
class CronFilterWithFeatureArgs:
    filter: FilterFunction
    feature_args: List[FeatureType]


class ResolverParsed(Protocol):
    function_definition: str
    filename: str
    fqn: str
    doc: Optional[str]
    inputs: Sequence[QueryInputType]
    output: FeatureSetType
    raw_fn: Callable
    fn: Callable
    environment: Optional[List[str]]
    tags: Optional[List[str]]
    machine_type: Optional[MachineType]
    when: Optional[FilterParsed]

    registry: "List[ResolverParsed]" = []

    def short_name(self) -> str:
        return self.fqn.split(".")[-1]

    def __hash__(self):
        return hash(self.fqn)

    @property
    def root_ns(self):
        """The @features cls namespace for which this resolver is rooted."""
        return get_unique_item([x.root_namespace() for x in self.inputs], name="root namespace")


class OnlineResolverParsed(ResolverParsed):
    cron: Union[ScheduleOptions, Cron]

    def __eq__(self, other):
        return isinstance(other, OnlineResolverParsed) and self.fqn == other.fqn

    def __hash__(self):
        return hash(self.fqn)

    def __init__(
        self,
        function_definition: str,
        filename: str,
        fqn: str,
        doc: Optional[str],
        inputs: List[QueryInputType],
        output: FeatureSetType,
        raw_fn: Callable,
        fn: Callable,
        environment: Optional[List[str]],
        tags: Optional[List[str]],
        cron: Union[ScheduleOptions, Cron],
        cron_filter: Optional[CronFilterWithFeatureArgs],
        machine_type: Optional[MachineType],
        when: Optional[FilterParsed],
    ):
        super(OnlineResolverParsed, self).__init__()
        self.filename = filename
        self.function_definition = function_definition
        self.fqn = fqn
        self.inputs = inputs
        self.output = output
        self.fn = fn
        self.raw_fn = raw_fn
        self.environment = environment
        self.tags = tags
        self.cron = cron
        self.cron_filter = cron_filter
        self.doc = doc
        self.machine_type = machine_type
        self.when = when

    def __repr__(self):
        return f"""OnlineResolverParsed(
  name={self.fqn},
  inputs={self.inputs}
  output={self.output}
)"""


class SinkResolverParsed:
    def __init__(
        self,
        function_definition: str,
        fqn: str,
        filename: str,
        doc: Optional[str],
        inputs: List[FeatureType],
        raw_fn: Callable,
        environment: Optional[List[str]],
        tags: Optional[List[str]],
        machine_type: Optional[MachineType],
        buffer_size: Optional[int],
        debounce: Optional[Duration],
        max_delay: Optional[Duration],
        upsert: Optional[bool],
    ):
        self.filename = filename
        self.function_definition = function_definition
        self.fqn = fqn
        self.inputs = inputs
        self.raw_fn = raw_fn
        self.environment = environment
        self.tags = tags
        self.doc = doc
        self.machine_type = machine_type
        self.buffer_size = buffer_size
        self.debounce = debounce
        self.max_delay = max_delay
        self.upsert = upsert

    def __repr__(self):
        return f"""SinkResolverParsed(
  name={self.fqn},
  inputs={self.inputs}
)"""

    def __eq__(self, other):
        return isinstance(other, SinkResolverParsed) and self.fqn == other.fqn

    def __hash__(self):
        return hash(self.fqn)


class OfflineResolverParsed(ResolverParsed):
    def __eq__(self, other):
        return isinstance(other, OfflineResolverParsed) and self.fqn == other.fqn

    def __hash__(self):
        return hash(self.fqn)

    def __init__(
        self,
        function_definition: str,
        fqn: str,
        filename: str,
        doc: Optional[str],
        inputs: List[QueryInputType],
        output: FeatureSetType,
        raw_fn: Callable,
        fn: Callable,
        environment: Optional[List[str]],
        tags: Optional[List[str]],
        cron: Union[ScheduleOptions, Cron],
        cron_filter: Optional[CronFilterWithFeatureArgs],
        machine_type: Optional[MachineType],
    ):
        self.filename = filename
        self.function_definition = function_definition
        self.fqn = fqn
        self.inputs = inputs
        self.output = output
        self.fn = fn
        self.raw_fn = raw_fn
        self.environment = environment
        self.tags = tags
        self.cron = cron
        self.cron_filter = cron_filter
        self.doc = doc
        self.machine_type = machine_type
        self.when = None

    def __repr__(self):
        return f"""OfflineResolverParsed(
  name={self.fqn},
  inputs={self.inputs}
  output={self.output}
)"""



from typing import Mapping, Optional, Union

import sqlalchemy.sql.functions
from sqlalchemy.orm import InstrumentedAttribute

from chalk.features import Feature, Features
from chalk.features.feature import FeatureWrapper, unwrap_feature
from chalk.sql.base.protocols import (
    BaseSQLSourceProtocol,
    ChalkQueryProtocol,
    DBSessionMakerProtocol,
    StringChalkQueryProtocol,
)
from chalk.sql.base.session import DBSessionMaker
from chalk.sql.integrations.chalk_query import ChalkQuery, StringChalkQuery


class BaseSQLSource(BaseSQLSourceProtocol):
    def __init__(self, session_maker: Optional[DBSessionMaker] = None):
        self._session_maker = session_maker or DBSessionMaker()
        self._incremental_settings = None

    def set_session_maker(self, maker: DBSessionMakerProtocol) -> None:
        self._session_maker = maker

    def query_string(
        self,
        query: str,
        fields: Mapping[str, Union[Feature, FeatureWrapper, str]],
        args: Optional[Mapping[str, str]] = None,
    ) -> StringChalkQueryProtocol:
        session = self._session_maker.get_session(self)
        fields = {f: unwrap_feature(v) if isinstance(v, FeatureWrapper) else v for (f, v) in fields.items()}

        return StringChalkQuery(session=session, source=self, query=query, fields=fields, args=args)

    def query(self, *entities, **kwargs) -> ChalkQueryProtocol:
        targets = []
        features = []
        for e in entities:
            if isinstance(e, Features):
                for f in e.features:
                    assert isinstance(f, Feature), f"Feature {f} must inherit from Feature"
                    assert f.name is not None
                    feature_value = getattr(e, f.name)
                    if isinstance(feature_value, InstrumentedAttribute):
                        features.append(f)
                        targets.append(feature_value.label(f.fqn))
                    elif isinstance(feature_value, sqlalchemy.sql.functions.GenericFunction):
                        features.append(f)
                        targets.append(feature_value.label(f.fqn))
            else:
                targets.append(e)
        session = self._session_maker.get_session(self)
        session.update_query(lambda x: x.query(*targets, **kwargs))

        return ChalkQuery(
            features=features,
            session=session,
            source=self,
        )



import os
import re
from codecs import open
from os import path

from setuptools import find_packages, setup

COMPANY_NAME = "chalk"
NAME = "chalk"
DESCRIPTION = f"Python SDK for Chalk"

REQUIRED = [
    "Click==8.*",
    "JSON-log-formatter",
    "aiosqlite",
    "dataclasses-json",
    "gitignore-parser>=0.0.8",
    "log-with-context",
    "mypy==0.931",
    "numpy",
    "pandas",
    "pandas-stubs",
    "pendulum",
    "psycopg2",
    "pydantic>=1.0.0",
    "pyyaml",
    "requests",
    "sqlalchemy>=1.4.26",
    "varname==0.9.0",
    "executing",
    "pure_eval",
    "cattrs==22.1.0",  # required to serialize dataclasses
]

repo_root = path.abspath(path.dirname(__file__))
LONG_DESCRIPTION = open(os.path.join(repo_root, "README.md"), "r").read()
VERSIONFILE = os.path.join(repo_root, "_version.py")
mo = re.search(
    r"^__version__ = ['\"]([^'\"]*)['\"]",
    open(VERSIONFILE, "rt").read(),
    re.M,
)

assert mo is not None, f"Unable to find version string in {str(VERSIONFILE)}"

version = mo.group(1)

setup(
    version=version,
    name="chalkpy",
    author="Chalk AI, Inc.",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    python_requires=">=3.8.0",
    url="https://chalk.ai",
    packages=find_packages(exclude=("tests",)),
    install_requires=REQUIRED,
    include_package_data=True,
    package_data={"chalk": ["py.typed"]},
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    entry_points={"console_scripts": [f"chalkpy=chalk.cli:cli"]},
    setup_requires=[
        "setuptools_scm",
    ],
)



from typing import Any, Mapping, Optional

from sqlalchemy.engine.url import URL

from chalk.integrations.named import load_integration_variable
from chalk.sql.base.sql_source import BaseSQLSource


class PostgreSQLSource(BaseSQLSource):
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[str] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        name: Optional[str] = None,
    ):
        self.host = host or load_integration_variable(integration_name=name, name="PGHOST")
        self.port = port or load_integration_variable(integration_name=name, name="PGPORT")
        self.db = db or load_integration_variable(integration_name=name, name="PGDATABASE")
        self.user = user or load_integration_variable(integration_name=name, name="PGUSER")
        self.password = password or load_integration_variable(integration_name=name, name="PGPASSWORD")

        super(PostgreSQLSource, self).__init__()

    def engine_args(self) -> Mapping[str, Any]:
        return dict(
            pool_size=20,
            max_overflow=60,
            # Trying to fix mysterious dead connection issue
            connect_args={
                "keepalives": 1,
                "keepalives_idle": 30,
                "keepalives_interval": 10,
                "keepalives_count": 5,
            },
        )

    def local_engine_url(self) -> str:
        return URL.create(
            drivername="postgresql",
            username=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.db,
        )



import unittest
from datetime import datetime

from chalk.features import DataFrame, Feature, Features, features, has_many, has_one, online
from chalk.features.feature import unwrap_feature
from chalk.features.feature_set import feature_time


@features
class ChildFS:
    parent_id: str
    parents: "DataFrame[ParentFS]"
    single_parent: "SingleParentFS"


@features
class SingleChildFS:
    parent_id: str
    parent: "DataFrame[ParentFS]"


@features
class SingleParentFS:
    id: str
    children: DataFrame[ChildFS] = has_many(lambda: ChildFS.parent_id == SingleParentFS.id)


@features
class ParentFS:
    id: str
    children: "DataFrame[ChildFS]" = has_many(lambda: ChildFS.parent_id == ParentFS.id)
    single_child: "SingleChildFS" = has_one(lambda: ParentFS.id == SingleChildFS.parent_id)
    ts: datetime = feature_time()


@online
def get_parent_time(p: ParentFS.id) -> Features[ParentFS.ts]:
    thing = ParentFS(ts=datetime.now())
    return thing


class BackPopulationOfFeaturesTestCase(unittest.TestCase):
    def test_one_to_one(self):
        """
        FeatureSet              -> FeatureSet
        SingleParentFS.children -> ChildFS.single_parent
        """
        parent_from_features: Feature = next((x for x in ChildFS.features if x.is_has_one))
        parent_from_attr = unwrap_feature(ChildFS.single_parent)
        self.assertEqual(parent_from_features, parent_from_attr)
        self.assertEqual(parent_from_features.name, "single_parent")
        self.assertEqual(parent_from_features.namespace, "child_fs")
        self.assertTrue(parent_from_features.is_has_one)
        assert parent_from_features.typ is not None
        self.assertEqual(parent_from_features.typ.underlying, SingleParentFS)
        self.assertEqual(parent_from_features.join, unwrap_feature(SingleParentFS.children).join)
        self.assertEqual(
            len(unwrap_feature(SingleParentFS.children).typ.parsed_annotation.columns),
            1,
        )

    def test_one_to_many(self):
        """
        FeatureSet            -> DataFrame
        ParentFS.single_child -> SingleChildFS.parent
        """
        parent_from_features: Feature = next((x for x in SingleChildFS.features if x.is_has_many))
        parent_from_attr = unwrap_feature(SingleChildFS.parent)
        self.assertEqual(parent_from_features, parent_from_attr)
        self.assertEqual(parent_from_features.name, "parent")
        self.assertEqual(parent_from_features.namespace, "single_child_fs")
        self.assertTrue(parent_from_features.is_has_many)
        assert parent_from_features.typ is not None
        self.assertTrue(issubclass(parent_from_features.typ.parsed_annotation, DataFrame))
        assert issubclass(parent_from_features.typ.parsed_annotation, DataFrame)  # for pylance
        self.assertEqual(parent_from_features.typ.parsed_annotation.references_feature_set, ParentFS)
        self.assertEqual(parent_from_features.join, unwrap_feature(ParentFS.single_child).join)

    def test_many_to_one(self):
        """
        DataFrame               -> FeatureSet
        SingleParentFS.children -> ChildFS.single_parent
        """
        parent_from_features = next((x for x in ChildFS.features if x.is_has_one))
        parent_from_attr = unwrap_feature(ChildFS.single_parent)
        self.assertEqual(parent_from_features, parent_from_attr)
        self.assertEqual(parent_from_features.name, "single_parent")
        self.assertEqual(parent_from_features.namespace, "child_fs")
        self.assertTrue(parent_from_features.is_has_one)
        self.assertEqual(parent_from_features.typ.underlying, SingleParentFS)
        self.assertEqual(parent_from_features.join, unwrap_feature(SingleParentFS.children).join)

    def test_many_to_many(self):
        """
        DataFrame         -> DataFrame
        ParentFS.children -> ChildFS.parents
        """
        parents_from_features: Feature = next((x for x in ChildFS.features if x.is_has_many))
        parents_from_attr = unwrap_feature(ChildFS.parents)
        self.assertEqual(parents_from_features, parents_from_attr)
        self.assertEqual(parents_from_features.name, "parents")
        self.assertEqual(parents_from_features.namespace, "child_fs")
        self.assertTrue(parents_from_features.is_has_many)
        assert parents_from_features.typ is not None
        assert isinstance(parents_from_features.typ.parsed_annotation, type)  # for pylance
        self.assertTrue(issubclass(parents_from_features.typ.parsed_annotation, DataFrame))
        assert issubclass(parents_from_features.typ.parsed_annotation, DataFrame)  # for pylance
        self.assertEqual(parents_from_features.typ.parsed_annotation.references_feature_set, ParentFS)
        self.assertEqual(parents_from_features.join, unwrap_feature(ParentFS.children).join)


if __name__ == "__main__":
    unittest.main()



import re


def to_snake_case(name: str):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    name = re.sub("__([A-Z])", r"_\1", name)
    name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", name)
    return name.lower()



import dataclasses
import inspect
from typing import Any, Callable, List, Optional, Union

from chalk.features.resolver import MachineType
from chalk.streams.StreamSource import StreamSource
from chalk.streams.StreamUpdate import StreamUpdate


class StreamResolver:
    registry = []

    def __init__(
        self,
        function_definition: str,
        fqn: str,
        filename: str,
        doc: Optional[str],
        source: StreamSource,
        fn: Callable,
        environment: Optional[Union[List[str], str]],
        machine_type: Optional[MachineType],
    ):
        super(StreamResolver, self).__init__()
        self.function_definition = function_definition
        self.fqn = fqn
        self.filename = filename
        self.source = source
        self.fn = fn
        self.environment = environment
        self.doc = doc
        self.machine_type = machine_type

    def __eq__(self, other):
        return isinstance(other, StreamResolver) and self.fqn == other.fqn

    def __hash__(self):
        return hash(self.fqn)

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    def __repr__(self):
        return f"StreamResolver(name={self.fqn})"


@dataclasses.dataclass
class StreamResolverParseResult:
    fqn: str
    source: StreamSource
    function_definition: str
    function: Callable
    doc: Optional[str]


def _parse_function(fn: Callable) -> StreamResolverParseResult:
    fqn = f"{fn.__module__}.{fn.__name__}"
    sig = inspect.signature(fn)

    function_definition = inspect.getsource(fn)

    if len(sig.parameters.values()) != 1:
        raise ValueError(f"Expected exactly one parameter for stream: {fqn}")

    arg = list(sig.parameters.values())[0].annotation

    return StreamResolverParseResult(
        fqn=fqn,
        source=arg,
        function_definition=function_definition,
        function=fn,
        doc=fn.__doc__,
    )


def stream(
    fn: Optional[Callable[[Any], StreamUpdate]] = None,
    environment: Optional[Union[List[str], str]] = None,
    machine_type: Optional[MachineType] = None,
):
    caller_filename = inspect.stack()[1].filename

    def decorator(args, cf=caller_filename):
        parsed = _parse_function(args)

        resolver = StreamResolver(
            filename=cf,
            source=parsed.source,
            function_definition=parsed.function_definition,
            fqn=parsed.fqn,
            doc=fn.__doc__,
            fn=parsed.function,
            environment=environment,
            machine_type=machine_type,
        )
        StreamResolver.registry.append(resolver)
        return resolver

    return decorator(fn) if fn else decorator



import enum
import unittest
from datetime import date

import pandas

from chalk.client import _ChalkAPIClientImpl
from chalk.client.client_impl import _OfflineQueryResponse
from chalk.features.feature import Feature
from chalk.serialization.codec import FeatureCodec


class Gender(enum.Enum):
    M = "m"
    F = "f"
    X = "x"


class TestClientSerialization(unittest.TestCase):
    def test_deserialize_null_int_enum(self):
        client = _ChalkAPIClientImpl(
            client_id="dummy",
            client_secret="dummy",
            environment="dummy",
            api_server="dummy",
        )

        client._codec = FeatureCodec(
            fqn_to_feature={
                "user.id": Feature(name="id", namespace="user", typ=int),
                "user.gender": Feature(name="gender", namespace="user", typ=Gender),
            }
        )

        fixture_response = _OfflineQueryResponse(
            columns=["user.id", "user.gender"],
            output=[
                [1, None],
                ["m", "f"],
            ],
        )

        # this shouldn't throw
        client._decode_offline_response(offline_query_response=fixture_response)

    def test_deserialize_date(self):
        client = _ChalkAPIClientImpl(
            client_id="dummy",
            client_secret="dummy",
            environment="dummy",
            api_server="dummy",
        )

        client._codec = FeatureCodec(
            fqn_to_feature={"user.birthday": Feature(name="birthday", namespace="user", typ=date)}
        )

        fixture_response = _OfflineQueryResponse(
            columns=["user.birthday"],
            output=[
                ["2022-09-08"],
            ],
        )

        decoded_frame = client._decode_offline_response(offline_query_response=fixture_response)
        pandas.testing.assert_frame_equal(
            decoded_frame,
            pandas.DataFrame(data={"user.birthday": [date.fromisoformat("2022-09-08")]}),
        )



from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import LetterCase, dataclass_json


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertFeatureIdGQL:
    fqn: str
    name: str
    namespace: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertReferencePathComponentGQL:
    parent: UpsertFeatureIdGQL
    child: UpsertFeatureIdGQL
    parent_to_child_attribute_name: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertFilterGQL:
    lhs: UpsertFeatureIdGQL
    op: str
    rhs: UpsertFeatureIdGQL


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertDataFrameGQL:
    columns: Optional[List[UpsertFeatureIdGQL]] = None
    filters: Optional[List[UpsertFilterGQL]] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertFeatureReferenceGQL:
    underlying: UpsertFeatureIdGQL
    path: Optional[List[UpsertReferencePathComponentGQL]] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertHasOneKindGQL:
    join: UpsertFilterGQL


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertHasManyKindGQL:
    join: UpsertFilterGQL
    columns: Optional[List[UpsertFeatureIdGQL]] = None
    filters: Optional[List[UpsertFilterGQL]] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertScalarKindGQL:
    scalar_kind: str
    primary: bool
    base_classes: List[str]
    version: Optional[int]
    has_encoder_and_decoder: bool = False


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertFeatureTimeKindGQL:
    format: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertFeatureGQL:
    id: UpsertFeatureIdGQL

    scalar_kind: Optional[UpsertScalarKindGQL] = None
    has_many_kind: Optional[UpsertHasManyKindGQL] = None
    has_one_kind: Optional[UpsertHasOneKindGQL] = None
    feature_time_kind: Optional[UpsertFeatureTimeKindGQL] = None

    etl_offline_to_online: bool = False
    tags: Optional[List[str]] = None
    max_staleness: Optional[str] = None
    description: Optional[str] = None
    owner: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class KafkaConsumerConfigGQL:
    broker: List[str]
    topic: List[str]
    ssl_keystore_location: Optional[str]
    client_id_prefix: Optional[str]
    group_id_prefix: Optional[str]
    topic_metadata_refresh_interval_ms: Optional[int]
    security_protocol: Optional[str]


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertStreamResolverGQL:
    fqn: str
    kind: str
    function_definition: str
    config: KafkaConsumerConfigGQL
    environment: Optional[List[str]] = None
    doc: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertResolverOutputGQL:
    features: Optional[List[UpsertFeatureIdGQL]] = None
    dataframes: Optional[List[UpsertDataFrameGQL]] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertResolverGQL:
    fqn: str
    kind: str
    function_definition: str
    output: UpsertResolverOutputGQL
    environment: Optional[List[str]] = None
    tags: Optional[List[str]] = None
    doc: Optional[str] = None
    cron: Optional[str] = None
    inputs: Optional[List[UpsertFeatureReferenceGQL]] = None
    machine_type: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertSinkResolverGQL:
    fqn: str
    function_definition: str
    environment: Optional[List[str]] = None
    tags: Optional[List[str]] = None
    doc: Optional[str] = None
    inputs: Optional[List[UpsertFeatureReferenceGQL]] = None
    machine_type: Optional[str] = None
    buffer_size: Optional[int] = None
    debounce: Optional[str] = None
    max_delay: Optional[str] = None
    upsert: Optional[bool] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class UpsertGraphGQL:
    resolvers: Optional[List[UpsertResolverGQL]] = None
    features: Optional[List[UpsertFeatureGQL]] = None
    streams: Optional[List[UpsertStreamResolverGQL]] = None
